#include <libleaftag/leaftag.h>
#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef struct
{
	int (*func)(int argc, char **argv);
	const char *progname;
	const char *param;

} Command;

static GList *
explode_tags_arg(const char *arg)
{
	GList *tag_names = NULL;
	gchar **strs;
	gchar **str;

	strs = g_strsplit(arg, ",", 0);

	for (str = strs; *str != NULL; str++)
		tag_names = g_list_prepend(tag_names, g_strdup(*str));

	g_strfreev(strs);

	return tag_names;
}

static int
command_export(int argc, char **argv)
{
	GList *tags, *sources, *l;

	if (argc != 0)
	{
		printf("Usage: tagutils export\n");
		return 1;
	}

	tags = lt_get_all_tags();

	for (l = tags; l != NULL; l = l->next)
	{
		LtTag *tag = LT_TAG(l->data);
		const char *desc;
		const char *image;

		printf("%%tag\tname: %s\thidden: %s",
			   lt_tag_get_name(tag),
			   lt_tag_get_hidden(tag) ? "true" : "false");

		if ((desc = lt_tag_get_description(tag)) != NULL)
		{
			char *temp = g_strescape(desc, NULL);
			printf("\tdescription: %s", temp);
			g_free(temp);
		}

		if ((image = lt_tag_get_image(tag)) != NULL)
		{
			char *temp = g_strescape(image, NULL);
			printf("\timage: %s", temp);
			g_free(temp);
		}

		printf("\n");
	}

	lt_free_object_list(tags);

	sources = lt_get_all_sources();

	for (l = sources; l != NULL; l = l->next)
	{
		LtSource *source = LT_SOURCE(l->data);
		char *uri_esc = g_strescape(lt_source_get_uri(source), NULL);
		GList *l2;

		printf("%%source\turi: %s\ttags: ", uri_esc);
		g_free(uri_esc);

		tags = lt_source_get_tags(source);

		for (l2 = tags; l2 != NULL; l2 = l2->next)
		{
			LtTag *tag = LT_TAG(l2->data);

			if (l2 != tags)
				printf(",");

			printf("%s", lt_tag_get_name(tag));
		}

		lt_free_object_list(tags);

		printf("\n");
	}

	lt_free_object_list(sources);

	return 0;
}

/*
 * TODO: Write this file to handle large files better.
 */
static int
command_import(int argc, char **argv)
{
	GError *error = NULL;
	gchar *contents;
	gchar **lines, **line;
	gsize len;
	GList *tags = NULL, *l;

	if (argc != 1)
	{
		printf("Usage: tagutils import filename\n");
		return 1;
	}

	if (!g_file_get_contents(argv[0], &contents, &len, &error))
	{
		fprintf(stderr, "Unable to import file: %s\n", error->message);
		g_error_free(error);
		return 1;
	}

	lines = g_strsplit(contents, "\n", 0);
	g_free(contents);

	for (line = lines; *line != NULL; line++)
	{
		gchar **parts, **part;
		GHashTable *data = g_hash_table_new_full(g_str_hash, g_str_equal,
												 g_free, g_free);

		if (**line == '\0')
			continue;

		parts = g_strsplit(*line, "\t", 0);

		for (part = parts + 1; *part != NULL; part++)
		{
			const char *key;
			const char *value = NULL;
			char *c;

			key = *part;

			if ((c = strchr(*part, ' ')) != NULL)
			{
				*(c - 1) = '\0'; /* chop off the ':' from the key */
				*c = '\0';
				value = c + 1;
			}

			if (value != NULL)
				g_hash_table_insert(data, g_strdup(key), g_strdup(value));
		}

		if (!strcmp(parts[0], "%tag"))
			tags = g_list_prepend(tags, data);
		else
		{
			if (!strcmp(parts[0], "%source"))
			{
				char *uri = g_hash_table_lookup(data, "uri");
				char *tags = g_hash_table_lookup(data, "tags");

				if (uri != NULL && tags != NULL)
				{
					LtSource *source = lt_create_source(uri);
					GList *tag_names = explode_tags_arg(tags);
					lt_source_tag_with_names(source, tag_names);
					g_list_foreach(tag_names, (GFunc)g_free, NULL);
					g_list_free(tag_names);
					g_object_unref(G_OBJECT(source));
				}
			}

			g_hash_table_destroy(data);
		}

		g_strfreev(parts);
	}

	g_strfreev(lines);

	for (l = tags; l != NULL; l = l->next)
	{
		GHashTable *data = (GHashTable *)l->data;
		LtTag *tag;
		char *name;
		char *hidden;
		char *desc;
		char *image;

		if ((name = g_hash_table_lookup(data, "name")) == NULL)
			continue;

		if ((tag = lt_tag_lookup(name)) == NULL)
			continue;

		hidden = g_hash_table_lookup(data, "hidden");

		if (!strcmp(hidden, "true"))
			lt_tag_set_hidden(tag, TRUE);

		if ((image = g_hash_table_lookup(data, "image")) != NULL)
		{
			char *temp = g_strcompress(image);
			lt_tag_set_image(tag, temp);
			g_free(temp);
		}

		if ((desc = g_hash_table_lookup(data, "description")) != NULL)
		{
			char *temp = g_strcompress(desc);
			lt_tag_set_description(tag, temp);
			g_free(temp);
		}

		g_object_unref(G_OBJECT(tag));
	}

	g_list_foreach(tags, (GFunc)g_hash_table_destroy, NULL);
	g_list_free(tags);

	return 0;
}

static void
command_ls_help(void)
{
	printf("Usage: tagutils ls tag[,...] [schema]\n");
}

static int
command_ls(int argc, char **argv)
{
	GList *tag_names, *l, *sources;
	char *schema = NULL;

	if (argc == 0 || argc > 2)
	{
		command_ls_help();
		return 1;
	}

	if (argc == 2)
		schema = argv[1];

	tag_names = explode_tags_arg(argv[0]);
	sources = lt_get_sources_with_tag_names(tag_names, schema);
	g_list_foreach(tag_names, (GFunc)g_free, NULL);
	g_list_free(tag_names);

	for (l = sources; l != NULL; l = l->next)
	{
		LtSource *source = LT_SOURCE(l->data);

		if (!strcmp(lt_source_get_schema(source), "file"))
		{
			const char *filename = lt_source_get_filename(source);

			if (g_file_test(filename, G_FILE_TEST_EXISTS))
				printf("%s\n", filename);
		}
		else
		{
			printf("%s\n", lt_source_get_uri(source));
		}
	}

	lt_free_object_list(sources);

	return 0;
}

static void
command_tag_help(void)
{
	printf("Usage: tagutils tag [-r] tag[,...] filename|uri...\n");
	printf("\n");
	printf("Parameters:\n");
	printf("    -r    - Tags any specified folders recursively\n");
}

static void
tag_recursive(const char *dir, GList *tag_names)
{
	GError *error = NULL;
	GDir *dirh;
	const char *filename;

	if ((dirh = g_dir_open(dir, 0, &error)) == NULL)
	{
		fprintf(stderr, "Error opening %s: %s\n", dir, error->message);
		g_error_free(error);
		return;
	}

	while ((filename = g_dir_read_name(dirh)) != NULL)
	{
		LtSource *source;
		char *path = g_build_path(G_DIR_SEPARATOR_S, dir, filename, NULL);

		source = lt_create_source(path);
		lt_source_tag_with_names(source, tag_names);
		g_object_unref(source);

		if (g_file_test(path, G_FILE_TEST_IS_DIR))
			tag_recursive(path, tag_names);

		g_free(path);
	}

	g_dir_close(dirh);
}

static int
command_tag(int argc, char **argv)
{
	int argi = 0;
	int i;
	gboolean recursive = FALSE;
	GList *tag_names;

	if (argc < 2)
	{
		command_tag_help();
		return 1;
	}

	if (!strcmp(argv[argi], "-r"))
	{
		recursive = TRUE;
		argi++;
	}

	tag_names = explode_tags_arg(argv[argi]);
	argi++;

	for (i = argi; i < argc; i++)
	{
		LtSource *source = lt_create_source(argv[i]);
		gboolean is_file = !strcmp(lt_source_get_schema(source), "file");

		if (is_file && !g_file_test(lt_source_get_filename(source),
									G_FILE_TEST_EXISTS))
		{
			fprintf(stderr, "%s does not exist\n",
					lt_source_get_filename(source));
		}
		else
		{
			lt_source_tag_with_names(source, tag_names);

			if (recursive && is_file &&
				g_file_test(lt_source_get_filename(source),
							G_FILE_TEST_IS_DIR))
			{
				tag_recursive(lt_source_get_filename(source), tag_names);
			}
		}

		g_object_unref(source);
	}

	g_list_foreach(tag_names, (GFunc)g_free, NULL);
	g_list_free(tag_names);

	return 0;
}

static void
command_tagprop_help(void)
{
    printf("Usage: tagutils tagprop [filename]\n");
    printf("\n");
    printf("Commands:\n");
    printf("    set    - sets a property on a tag\n");
    printf("             tagprop set tag[,...] property value\n");
    printf("\n");
    printf("    unset  - unsets a property on a tag\n");
    printf("             tagprop unset tag[,...] property\n");
    printf("\n");
    printf("    get    - returns a property on a tag\n");
    printf("             tagprop get tag property\n");
    printf("\n");
    printf("    list   - displays a list of properties on a tag\n");
    printf("             tagprop list tag\n");
}

static void
validate_tag_property(const char *prop)
{
	if (strcmp(prop, "image") && strcmp(prop, "description") &&
		strcmp(prop, "hidden"))
	{
		fprintf(stderr, "Unknown property '%s'\n", prop);
		exit(1);
	}
}

static int
command_tagprop(int argc, char **argv)
{
	char *command;

	if (argc < 1)
		goto error;

	command = argv[0];

	if (!strcmp(command, "set"))
	{
		char *prop, *value;
		GList *tag_names, *tags, *l;

		if (argc != 4)
			goto error;

		prop = argv[2];
		validate_tag_property(prop);

		tag_names = explode_tags_arg(argv[1]);
		tags = lt_tag_lookup_many(tag_names);
		g_list_foreach(tag_names, (GFunc)g_free, NULL);
		g_list_free(tag_names);

		value = argv[3];

		for (l = tags; l != NULL; l = l->next)
		{
			LtTag *tag = LT_TAG(l->data);

			if (!strcmp(prop, "description"))
				lt_tag_set_description(tag, value);
			else if (!strcmp(prop, "image"))
				lt_tag_set_image(tag, value);
			else if (!strcmp(prop, "hidden"))
				lt_tag_set_hidden(tag, !strcasecmp(value, "true"));
			else
				g_assert_not_reached();
		}

		lt_free_object_list(tags);
	}
	else if (!strcmp(command, "get"))
	{
		const char *value;
		char *tagname, *prop;
		LtTag *tag;

		if (argc != 3)
			goto error;

		tagname = argv[1];
		prop = argv[2];
		validate_tag_property(prop);

		tag = lt_tag_lookup(tagname);

		if (tag == NULL)
		{
			fprintf(stderr, "The tag '%s' does not exist\n", tagname);
			return 1;
		}

		if (!strcmp(prop, "description"))
			value = lt_tag_get_description(tag);
		else if (!strcmp(prop, "image"))
			value = lt_tag_get_image(tag);
		else if (!strcmp(prop, "hidden"))
			value = lt_tag_get_hidden(tag) ? "true" : "false";
		else
			g_assert_not_reached();

		if (value != NULL)
			printf("%s\n", value);

		g_object_unref(G_OBJECT(tag));
	}
	else if (!strcmp(command, "unset"))
	{
		GList *tag_names, *tags, *l;
		char *prop;

		if (argc != 3)
			goto error;

		prop = argv[2];
		validate_tag_property(prop);

		tag_names = explode_tags_arg(argv[1]);
		tags = lt_tag_lookup_many(tag_names);
		g_list_foreach(tag_names, (GFunc)g_free, NULL);
		g_list_free(tag_names);

		for (l = tags; l != NULL; l = l->next)
		{
			LtTag *tag = LT_TAG(l->data);

			if (!strcmp(prop, "description"))
				lt_tag_set_description(tag, NULL);
			else if (!strcmp(prop, "image"))
				lt_tag_set_image(tag, NULL);
			else if (!strcmp(prop, "hidden"))
				lt_tag_set_hidden(tag, FALSE);
			else
				g_assert_not_reached();
		}

		lt_free_object_list(tags);
	}
	else if (!strcmp(command, "list"))
	{
		LtTag *tag;

		if (argc != 2)
			goto error;

		tag = lt_tag_lookup(argv[1]);

		if (tag == NULL)
		{
			fprintf(stderr, "Tag not found\n");
		}
		else
		{
			const char *desc = lt_tag_get_description(tag);
			const char *image = lt_tag_get_image(tag);

			printf("description: %s\n", desc == NULL ? "" : desc);
			printf("hidden: %s\n", lt_tag_get_hidden(tag) ? "true" : "false");
			printf("image: %s\n", image == NULL ? "" : image);
		}

		g_object_unref(G_OBJECT(tag));
	}
	else
	{
		fprintf(stderr, "Invalid command '%s'\n", command);
		goto error;
	}

	return 0;

error:
	command_tagprop_help();
	return 1;
}

static void
command_tags_help(void)
{
	printf("Usage: tagutils tags [filename|uri]\n");
}

static int
command_tags(int argc, char **argv)
{
	GList *tags = NULL, *l;

	if (argc > 1)
	{
		command_tags_help();
		return 1;
	}
	else if (argc == 1)
	{
		LtSource *source = lt_source_lookup(argv[0]);

		if (source != NULL)
			tags = lt_source_get_tags(source);
	}
	else
	{
		tags = lt_get_all_tags();
	}

	for (l = tags; l != NULL; l = l->next)
		printf("%s\n", lt_tag_get_name(LT_TAG(l->data)));

	lt_free_object_list(tags);

	return 0;
}

static void
command_untag_help(void)
{
	printf("Usage: tagutils untag [-r] tag[,...] filename|uri...\n");
	printf("\n");
	printf("Parameters:\n");
	printf("    -r    - Untags any specified folders recursively\n");
}

static void
untag_recursive(const char *dir, GList *tag_names)
{
	GError *error = NULL;
	GDir *dirh;
	const char *filename;

	if ((dirh = g_dir_open(dir, 0, &error)) == NULL)
	{
		fprintf(stderr, "Error opening %s: %s\n", dir, error->message);
		g_error_free(error);
		return;
	}

	while ((filename = g_dir_read_name(dirh)) != NULL)
	{
		LtSource *source;
		char *path = g_build_path(G_DIR_SEPARATOR_S, dir, filename, NULL);

		source = lt_source_lookup(path);

		if (source != NULL)
		{
			lt_source_untag_with_names(source, tag_names);
			g_object_unref(source);
		}

		if (g_file_test(path, G_FILE_TEST_IS_DIR))
			untag_recursive(path, tag_names);

		g_free(path);
	}

	g_dir_close(dirh);
}

static int
command_untag(int argc, char **argv)
{
	int argi = 0;
	int i;
	gboolean recursive = FALSE;
	GList *tag_names;

	if (argc < 2)
	{
		command_untag_help();
		return 1;
	}

	if (!strcmp(argv[argi], "-r"))
	{
		recursive = TRUE;
		argi++;
	}

	tag_names = explode_tags_arg(argv[argi]);
	argi++;

	for (i = argi; i < argc; i++)
	{
		LtSource *source = lt_source_lookup(argv[i]);
		gboolean is_file;

		if (source == NULL)
			continue;

		is_file = !strcmp(lt_source_get_schema(source), "file");

		if (is_file && !g_file_test(lt_source_get_filename(source),
									G_FILE_TEST_EXISTS))
		{
			fprintf(stderr, "%s does not exist\n",
					lt_source_get_filename(source));
		}
		else
		{
			lt_source_untag_with_names(source, tag_names);

			if (recursive && is_file &&
				g_file_test(lt_source_get_filename(source),
							G_FILE_TEST_IS_DIR))
			{
				untag_recursive(lt_source_get_filename(source), tag_names);
			}
		}

		g_object_unref(G_OBJECT(source));
	}

	g_list_foreach(tag_names, (GFunc)g_free, NULL);
	g_list_free(tag_names);

	return 0;
}

static void
help(void)
{
    printf("Usage: tagutils command [options...]\n");
    printf("\n");
    printf("Commands:\n");
    printf("    ls      - list files, folders and URIs with the given tag\n");
    printf("    tag     - tag a file or URI\n");
    printf("    tagprop - manage properties on a tag\n");
    printf("    tags    - list all tags or just tags for a file or URI\n");
    printf("    untag   - untag a file or URI\n");
    printf("\n");
    printf("Use 'tagutils command --help' for more information on a command\n");
}

static const Command commands[] =
{
	{ command_export,  NULL,      "export"  },
	{ command_import,  NULL,      "import"  },
	{ command_ls,      "tagls",   "ls"      },
	{ command_tag,     "tag",     "tag"     },
	{ command_tagprop, "tagprop", "tagprop" },
	{ command_tags,    "tags",    "tags"    },
	{ command_untag,   "untag",   "untag"   },
	{ NULL, NULL, NULL }
};

int
main(int argc, char **argv)
{
	const Command *command;
	char *cmdarg = g_path_get_basename(argv[0]);
	gboolean use_progname;
	int ret = 0;

	g_type_init();

	if (!strcmp(cmdarg, "tagutils"))
	{
		if (argc == 1)
		{
			fprintf(stderr, "Missing parameters passed to tagutils\n");
			goto error;
		}

		g_free(cmdarg);
		cmdarg = g_strdup(argv[1]);

		if (!strcmp(cmdarg, "-h") || !strcmp(cmdarg, "--help") ||
			!strcmp(cmdarg, "help"))
		{
			goto error;
		}

		use_progname = FALSE;
		argv += 2;
		argc -= 2;
	}
	else
	{
		use_progname = TRUE;
		argv++;
		argc--;
	}

	for (command = commands; command->func != NULL; command++)
	{
		const char *temp = (use_progname ? command->progname : command->param);

		if (temp != NULL && !strcmp(cmdarg, temp))
		{
			ret = command->func(argc, argv);
			goto exit;
		}
	}

	fprintf(stderr, "Invalid command specified\n");

error:
	help();
	ret = 1;

exit:
	exit(ret);
}
